/******************************************************
 *
 * Name:         examples.sql
 *     
 * Design Phase:
 *     Author:   John Miner
 *     Date:     Feb 2012
 *     Purpose:  Demonstrate the following tasks
 *               
 *               1 - Verify database
 *               2 - Full database backup
 *               3 - Differential database backup
 *               4 - Log backup
 *               5 - Generate prime #'s between backups
 *               6 - Cause data deletion
 *               7 - Tail log backup
 *               8 - Perform a correct order restore
 *               9 - Maintain msdb
 *               10 - Maintain database indexes
 *               11 - Maintain database statistics
 * 
 ******************************************************/


/*  
	Create a database to hold the prime numbers
*/


-- Which database to use.
USE [master]
GO

-- Delete existing databases.
IF  EXISTS (SELECT name FROM sys.databases WHERE name = N'BASIC_MATH')
    DROP DATABASE BASIC_MATH
GO

-- Add new databases based on model
CREATE DATABASE BASIC_MATH;
GO


/*  
	Create a table to hold the primes
*/

-- Which database to use.
USE [BASIC_MATH]
GO

-- Delete existing table
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'[DBO].[PRIMES]') AND type in (N'U'))
DROP TABLE [DBO].[PRIMES]
GO

-- Add new table
CREATE TABLE [DBO].[PRIMES] (
	[MY_VALUE] [bigint] NOT NULL
) ON [PRIMARY]
GO

-- Delete existing index
IF EXISTS (SELECT name FROM sys.indexes WHERE name = N'IDX_PRIMES')
    DROP INDEX IDX_PRIMES ON [DBO].[PRIMES];
GO

-- Add new index
CREATE CLUSTERED INDEX IDX_PRIMES 
    ON [DBO].[PRIMES] ([MY_VALUE]); 
GO


/*  
	Create a procedure to determine if number is prime
*/


-- Which database to use.
USE [BASIC_MATH]
GO

-- Delete existing procedure
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'[dbo].[IS_PRIME]') AND type in (N'P', N'PC'))
DROP PROCEDURE [dbo].[IS_PRIME]
GO

-- Create the stored procedure from scratch
CREATE PROCEDURE IS_PRIME
    @VAR_NUM2 BIGINT
AS
BEGIN
    -- LOCAL VARIABLES
    DECLARE @VAR_CNT2 BIGINT;
    DECLARE @VAR_MAX2 BIGINT;

    -- SET UP COUNTERS    
    SELECT @VAR_CNT2 = 2;
    SELECT @VAR_MAX2 = SQRT(@VAR_NUM2) + 1;

    -- TRIAL DIVISION 2 TO SQRT(X)
    WHILE (@VAR_CNT2 <= @VAR_MAX2)
    BEGIN
        -- NOT A PRIME NUMBER
        IF (@VAR_NUM2 % @VAR_CNT2) = 0
            RETURN 0;            

        -- INCREMENT COUNTER
        SELECT @VAR_CNT2 = @VAR_CNT2 + 1;
        
    END;

    -- NOT A PRIME NUMBER
    RETURN 1;
    
END
GO


/*  
	Create a procedure to store primes from x to y.
*/

-- Which database to use.
USE [BASIC_MATH]
GO

-- Delete existing procedure
IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'[dbo].[STORE_PRIMES]') AND type in (N'P', N'PC'))
DROP PROCEDURE [dbo].[STORE_PRIMES]
GO

-- Create the stored procedure from scratch
CREATE PROCEDURE STORE_PRIMES
    @VAR_ALPHA BIGINT,
    @VAR_OMEGA BIGINT
AS
BEGIN
    -- DECLARE VARIABLES
    DECLARE @VAR_CNT1 BIGINT;
    DECLARE @VAR_RET1 INT;
    
    -- SET VARIABLES
    SELECT @VAR_RET1 = 0;
    SELECT @VAR_CNT1 = @VAR_ALPHA;

    -- CHECK EACH NUMBER FOR PRIMENESS
    WHILE (@VAR_CNT1 <= @VAR_OMEGA)
    BEGIN
        -- ARE WE PRIME?
        EXEC @VAR_RET1 = DBO.IS_PRIME @VAR_CNT1;
        
        -- FOUND A PRIME
        IF (@VAR_RET1 = 1)
        BEGIN
            INSERT INTO [DBO].[PRIMES] (MY_VALUE) VALUES (@VAR_CNT1);
            PRINT @VAR_CNT1;
        END
        
        -- INCREMENT COU
        SELECT @VAR_CNT1 = @VAR_CNT1 + 1        
    END;
    
END
GO


--
-- Demonstrate Maintenance
--

-- CLEAR TABLE
TRUNCATE TABLE PRIMES

-- 167 PRIMES
EXEC DBO.STORE_PRIMES 2, 1000
GO

-- STEP 1 - NIGHTLY VERIFY DB
EXEC msdb.dbo.usp_Verify_Db_Integrity @VAR_NAME='BASIC_MATH', @VAR_NOINDEX=1
GO


-- STEP 2A - NIGHTLY FULL BACKUP (FORMAT)
EXEC msdb.dbo.usp_Backup_Db2File @VAR_PATH = 'C:\MSSQL\BACKUP\', @VAR_NAME='BASIC_MATH', @VAR_FORMAT = 1
GO

-- STEP 2B - VERIFY FULL BACKUP (NEW FILE)
EXEC msdb.dbo.usp_Verify_Backup4Db @VAR_PATH = 'C:\MSSQL\BACKUP\', @VAR_NAME='BASIC_MATH'
GO


-- STEP 3A - NIGHTLY LOG BACKUP (FORMAT)
EXEC msdb.dbo.usp_Backup_Log2File @VAR_PATH = 'C:\MSSQL\BACKUP\', @VAR_NAME='BASIC_MATH', @VAR_FORMAT = 1
GO

-- STEP 3B - VERIFY LOG BACKUP (NEW FILE)
EXEC msdb.dbo.usp_Verify_Backup4Log @VAR_PATH = 'C:\MSSQL\BACKUP\', @VAR_NAME='BASIC_MATH'
GO

  
-- 127 PRIMES
EXEC DBO.STORE_PRIMES 2000, 3000
GO

-- STEP 4A - DAILY DIFF BACKUP (FORMAT)
EXEC msdb.dbo.usp_Backup_Dif2File @VAR_PATH = 'C:\MSSQL\BACKUP\', @VAR_NAME='BASIC_MATH', @VAR_FORMAT = 1
GO

-- STEP 4B - VERIFY DIFF BACKUP (NEW FILE)
EXEC msdb.dbo.usp_Verify_Backup4Dif @VAR_PATH = 'C:\MSSQL\BACKUP\', @VAR_NAME='BASIC_MATH'
GO

-- 135 PRIMES
EXEC DBO.STORE_PRIMES 1001, 1999
GO


-- STEP 5A - DAILY LOG BACKUP (FORMAT)
EXEC msdb.dbo.usp_Backup_Log2File @VAR_PATH = 'C:\MSSQL\BACKUP\', @VAR_NAME='BASIC_MATH'
GO

-- STEP 5B - VERIFY LOG BACKUP (NEW FILE)
EXEC msdb.dbo.usp_Verify_Backup4Log @VAR_PATH = 'C:\MSSQL\BACKUP\', @VAR_NAME='BASIC_MATH'
GO


-- 120 PRIMES
EXEC DBO.STORE_PRIMES 3001, 3999
GO


-- STEP 6A - DAILY LOG BACKUP (FORMAT)
EXEC msdb.dbo.usp_Backup_Log2File @VAR_PATH = 'C:\MSSQL\BACKUP\', @VAR_NAME='BASIC_MATH'
GO

-- STEP 6B - VERIFY LOG BACKUP (NEW FILE)
EXEC msdb.dbo.usp_Verify_Backup4Log @VAR_PATH = 'C:\MSSQL\BACKUP\', @VAR_NAME='BASIC_MATH'
GO

-- 119 PRIMES
EXEC DBO.STORE_PRIMES 4001, 4999
GO

-- TOTAL OF 668 PRIMES
SELECT COUNT(*) FROM [BASIC_MATH].[dbo].[PRIMES]
GO

-- MIN = 2, MAX = 4999
SELECT MIN(MY_VALUE) AS MIN1, MAX(MY_VALUE) AS MAX1 FROM [BASIC_MATH].[dbo].[PRIMES]
GO


--
-- Damage caused by user, number divided by 5 leaves remainder of 1
--

-- Check algorithm
SELECT * FROM [BASIC_MATH].[dbo].[PRIMES] WHERE MY_VALUE % 5 = 1

-- Remove 163 random rows
BEGIN TRAN 
DELETE FROM [BASIC_MATH].[dbo].[PRIMES] WHERE MY_VALUE % 5 = 1
COMMIT


--
-- Demonstrate Restore
--

--
-- CHANGE DATE WHEN RESTORING (MON-SUN)
-- 
-- 1 - DO TAIL LOG BACKUP
-- 2 - RESTORE FULL
-- 3 - RESTORE DIFF
-- 4 - RESTORE LOG (#2)
-- 5 - RESTORE LOG (#3)
-- 6 - RESTORE TAIL

-- PERFORM TAIL BACKUP
DECLARE @FILE VARCHAR(256)
SET @FILE = 'C:\MSSQL\BACKUP\BASIC_MATH\TAIL_LOG.TRN'
BACKUP LOG BASIC_MATH
TO DISK = @FILE
WITH 
        FORMAT, 
	    INIT,  
	    NAME = 'Tail Log Backup', 
	    SKIP, 
	    NOREWIND, 
	    NOUNLOAD,  
	    STATS = 10,
        CONTINUE_AFTER_ERROR
GO

-- Can use NO_TRUNCATE as last resort
-- http://msdn.microsoft.com/en-us/library/ms179314.aspx

-- Switch to master
USE master
GO

-- START WITH FULL BACKUP
RESTORE DATABASE [BASIC_MATH] 
FROM DISK = N'C:\MSSQL\BACKUP\BASIC_MATH\BASIC_MATH_WED_FULL.BAK' 
WITH FILE = 1, REPLACE, NORECOVERY, NOUNLOAD, STATS = 10
GO

-- ADD DIFF BACKUP
RESTORE DATABASE [BASIC_MATH] 
FROM DISK = N'C:\MSSQL\BACKUP\BASIC_MATH\BASIC_MATH_WED_DIFF.BAK' 
WITH FILE = 1, NORECOVERY, NOUNLOAD, REPLACE, STATS = 10
GO

-- ADD LOG #2
RESTORE LOG [BASIC_MATH] 
FROM DISK = N'C:\MSSQL\BACKUP\BASIC_MATH\BASIC_MATH_WED_LOG.TRN' 
WITH FILE = 2, NORECOVERY, NOUNLOAD, STATS = 10
GO

-- ADD LOG #3
RESTORE LOG [BASIC_MATH] 
FROM DISK = N'C:\MSSQL\BACKUP\BASIC_MATH\BASIC_MATH_WED_LOG.TRN' 
WITH FILE = 3, NORECOVERY, NOUNLOAD, STATS = 10
GO

-- ADD TAIL LOG JUST BEFORE DAMAGE
RESTORE LOG [BASIC_MATH] 
FROM DISK = N'C:\MSSQL\BACKUP\BASIC_MATH\TAIL_LOG.TRN' 
WITH RECOVERY, STOPAT = '02-15-2012 3:58 PM';
GO

-- RECOVERED ALL 668 PRIMES
SELECT COUNT(*) FROM [BASIC_MATH].[dbo].[PRIMES]
GO


--
-- Demonstrate Maintenance
--


-- CLEAR OUT MSDB HISTORY > 30 DAYS
EXEC msdb.dbo.usp_Maintain_Msdb @VAR_DAYS = 30;
GO

-- MAINTAIN THE INDEXES
EXEC msdb.dbo.usp_Maintain_Indexes @database_name = 'BASIC_MATH';
GO

-- MAINTAIN THE STATISTICS
EXEC msdb.dbo.usp_Maintain_Statistics @VAR_NAME = 'BASIC_MATH';
GO
