/******************************************************
 *
 * Name:         usp_Maintain_Indexes.sql
 *     
 * Design Phase:
 *     Author:   John Miner
 *     Date:     11-14-2011
 *     Purpose:  Maintain the indexes on the tables.
 *     Revised:  Found better script to use.
 * 
 ******************************************************/


-- NOTE:
--     FOUND BETTER SCRIPT AT SQL SERVER PEDIA.
--     http://sqlserverpedia.com/w/index.php?title=Index_Maintenance&stable=1  


/*  
	Which database to use.
*/

USE msdb
GO


/*  
	Delete the existing table.
*/

If Exists(Select [object_id] From sys.tables Where name = N'tbl_Maintain_Indexes')
Begin
    drop table dbo.tbl_Maintain_Indexes
End
GO


/*  
	Create table to log index rebuilds
*/

If Not Exists(Select [object_id] From sys.tables Where name = N'tbl_Maintain_Indexes')
Begin    
    Create Table dbo.tbl_Maintain_Indexes
    (
          indexDefrag_id    int identity(1,1)   Not Null
        , databaseID        int                 Not Null
        , databaseName      nvarchar(128)       Not Null
        , objectID          int                 Not Null
        , objectName        nvarchar(128)       Not Null
        , indexID           int                 Not Null
        , indexName         nvarchar(128)       Not Null
        , partitionNumber   smallint            Not Null
        , fragmentation     float               Not Null
        , page_count        int                 Not Null
        , dateTimeStart     datetime            Not Null
        , durationSeconds   int                 Not Null
        Constraint PK_indexDefragLog Primary Key Clustered (indexDefrag_id)
    );

    -- Debug line
    Print 'dbo.tbl_Maintain_Indexes Table Created';    
End
GO


/*  
	Delete the existing stored procedure.
*/

IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'[dbo].[usp_Maintain_Indexes]') AND type in (N'P', N'PC'))
DROP PROCEDURE [dbo].[usp_Maintain_Indexes]
GO


/*  
	Create the new stored procedure.
*/


CREATE PROCEDURE [dbo].[usp_Maintain_Indexes]
      @reorganize_pt FLOAT = 10.0  
    , @rebuild_pt FLOAT = 30.0  
    , @report_only BIT = 0     
    , @database_name VARCHAR(512) = NULL
    , @table_name VARCHAR(4000) = Null 
    , @online_rebuild BIT = 1     
    , @MaxDop_Restriction TINYINT = Null
    , @print_commands BIT = 1    
    , @print_fragmentation BIT = 1          
    , @defrag_delay CHAR(8) = '00:00:05'
    , @scan_mode NVARCHAR(15) = N'SAMPLED'
    , @debug_mode BIT = 0
AS

/*
    CAUTION: TRANSACTION LOG SIZE MUST BE MONITORED CLOSELY WHEN DEFRAGMENTING.
 
      @reorganize_pt        defaulted to 10%, will not defrag if fragmentation 
                            is less than that
 
      @rebuild_pt           defaulted to 30% as recommended by Microsoft in BOL;
                            greater than 30% will result in rebuild instead
 
      @report_only          1 = execute the SQL generated by this proc; 
                            0 = print command only
 
      @database_name        Optional, specify specific database name to defrag;
                            If not specified, all non-system databases will
                            be defragged.
 
      @table_name           Specify if you only want to defrag indexes for a 
                            specific table, format = databaseName.schema.tableName;
                            if not specified, all tables will be defragged.
 
      @online_rebuild       1 = online rebuild; 
                            0 = offline rebuild
 
      @MaxDop_Restriction   Option to specify a processor limit for index rebuilds
 
      @print_commands       1 = print commands to screen; 
                            0 = do not print commands
 
      @print_fragmentation  1 = print fragmentation to screen;
                            0 = do not print fragmentation
 
      @defrag_delay         time to wait between defrag commands; gives the
                            server a little time to catch up 
      
      @scan_mode            scan level to be used with dm_db_index_physical_stats. 
                            Options are DEFAULT, NULL, LIMITED, SAMPLED, or 
                            DETAILED. The default (NULL) is LIMITED
 
      @debug_mode           1 = display debug comments; helps with troubleshooting
                            0 = do not display debug comments
 
*/																
 
SET NOCOUNT ON;
SET XACT_Abort ON;
SET Quoted_Identifier ON;
 
BEGIN
 
    -- Debug line
    IF (@debug_mode = 1 )
        RAISERROR('Executing usp_Maintain_Indexes ...', 0, 42) WITH NOWAIT;
 
    /* Declare our variables */
    DECLARE   @objectID             INT
            , @database_nameID      INT
            , @database_nameName    NVARCHAR(128)
            , @indexID              INT
            , @partitionCount       BIGINT
            , @schemaName           NVARCHAR(128)
            , @objectName           NVARCHAR(128)
            , @indexName            NVARCHAR(128)
            , @partitionNumber      SMALLINT
            , @partitions           SMALLINT
            , @fragmentation        FLOAT
            , @pageCount            INT
            , @sqlCommand           NVARCHAR(4000)
            , @rebuildCommand       NVARCHAR(200)
            , @dateTimeStart        DATETIME
            , @dateTimeEnd          DATETIME
            , @containsLOB          BIT
            , @editionCheck         BIT
            , @debugMessage         VARCHAR(128)
            , @updateSQL            NVARCHAR(4000)
            , @partitionSQL         NVARCHAR(4000)
            , @partitionSQL_Param   NVARCHAR(1000)
            , @LOB_SQL              NVARCHAR(4000)
            , @LOB_SQL_Param        NVARCHAR(1000);
 
    /*  Temporary table - indexes to defrag */
    CREATE TABLE #indexDefragList
    (
          databaseID        INT
        , databaseName      NVARCHAR(128)
        , objectID          INT
        , indexID           INT
        , partitionNumber   SMALLINT
        , fragmentation     FLOAT
        , page_count        INT
        , defragStatus      BIT
        , schemaName        NVARCHAR(128)   Null
        , objectName        NVARCHAR(128)   Null
        , indexName         NVARCHAR(128)   Null
    );
 
    /*  Temporary table - list of databases */
    CREATE TABLE #databaseList
    (
          databaseID        INT
        , databaseName      VARCHAR(128)
    );
 
    /*  Temporary table - list of processors */
    CREATE TABLE #processor 
    (
          [INDEX]           INT
        , Name              VARCHAR(128)
        , Internal_Value    INT
        , Character_Value   INT
    );
 

    -- Debug line
    IF (@debug_mode = 1)
        RAISERROR('Step 1 - Beginning validation of input parameters ...', 0, 42) WITH NOWAIT;
 
    /*  Validate input parameters  */
    IF (@reorganize_pt Not Between 0.00 And 100.0)
        SET @reorganize_pt = 5.0;
 
    IF (@rebuild_pt Not Between 0.00 And 100.0)
        SET @rebuild_pt = 30.0;
 
    IF (@defrag_delay Not Like '00:[0-5][0-9]:[0-5][0-9]')
        SET @defrag_delay = '00:00:05';
 
 
    /* Make sure we're not exceeding the number of processors we have available */
    INSERT INTO #processor
        EXECUTE XP_MSVER 'ProcessorCount';
 
    IF (@MaxDop_Restriction IS NOT NULL) And (@MaxDop_Restriction > (SELECT Internal_Value FROM #processor))
        SELECT @MaxDop_Restriction = Internal_Value FROM #processor;
 
    /* Check our server version; 1804890536 = Enterprise, 610778273 = Enterprise Evaluation, -2117995310 = Developer */
    IF (SELECT SERVERPROPERTY('EditionID')) In (1804890536, 610778273, -2117995310) 
        SET @editionCheck = 1 -- supports online rebuilds
    ELSE
        SET @editionCheck = 0; -- does not support online rebuilds
 

    -- Debug line
    IF (@debug_mode = 1)
        RAISERROR('Step 2 - Grabbing a list of our databases ...', 0, 42) WITH NOWAIT;
 
    /* Retrieve the list of databases to investigate (1 or N) */
    INSERT INTO #databaseList
        SELECT database_id, name
        FROM sys.databases
        WHERE 
            name = IsNull(@database_name, name)
            And database_id > 4 -- exclude system databases
            And [STATE] = 0; -- state must be ONLINE
 

    -- Debug line
    IF (@debug_mode = 1)
        RAISERROR('Step 3 - Examining databases; checking for fragmentation ...', 0, 42) WITH NOWAIT;
 
    /* Loop through our list of databases */
    WHILE (SELECT COUNT(*) FROM #databaseList) > 0
    BEGIN
 
        -- Get topmost database name
        SELECT TOP 1 
            @database_nameID = databaseID, 
            @debugMessage = 'Working on ' + DB_NAME(databaseID) + ' ...'
        FROM #databaseList;
 
        -- Debugging line
        IF (@debug_mode = 1)
            RAISERROR(@debugMessage, 0, 42) WITH NOWAIT;
 
       /* Determine which indexes to defrag using our user-defined parameters */
        INSERT INTO #indexDefragList
        SELECT
              database_id AS databaseID
            , QUOTENAME(DB_NAME(database_id)) AS 'databaseName'
            , [OBJECT_ID] AS objectID
            , index_id AS indexID
            , partition_number AS partitionNumber
            , avg_fragmentation_in_percent AS fragmentation
            , page_count 
            , 0 AS 'defragStatus' /* 0 = unprocessed, 1 = processed */
            , Null AS 'schemaName'
            , Null AS 'objectName'
            , Null AS 'indexName'
        FROM 
            sys.dm_db_index_physical_stats (@database_nameID, OBJECT_ID(@table_name), Null , Null, @scan_mode)
        WHERE 
            avg_fragmentation_in_percent >= @reorganize_pt 
            And index_id > 0 -- ignore heaps
            And page_count > 8 -- ignore objects with less than 1 extent
        OPTION (MAXDOP 1);
 
        -- Remove topmost datbase from list
        DELETE FROM #databaseList
        WHERE databaseID = @database_nameID;
 
    END
 
    -- Add index to temporary table
    CREATE CLUSTERED INDEX CIX_temp_indexDefragList
        ON #indexDefragList(databaseID, objectID, indexID, partitionNumber);
 
 
    -- Debug line
    SELECT @debugMessage = 'Step 4 - Looping through fragmented index list ...  There''s ' + CAST(COUNT(*) AS VARCHAR(10)) + ' indexes to defrag!'
    FROM #indexDefragList;   
    IF (@debug_mode = 1)
        RAISERROR(@debugMessage, 0, 42) WITH NOWAIT;
 
    /* Begin our loop for defragging */
    WHILE (SELECT COUNT(*) FROM #indexDefragList WHERE defragStatus = 0) > 0
    BEGIN
 
        -- Debug line
        IF (@debug_mode = 1)
            RAISERROR('Step 5A - Picking an index to beat into shape ... ', 0, 42) WITH NOWAIT;
 
        /* Grab the most fragmented index first to defrag */
        SELECT TOP 1 
              @objectID         = objectID
            , @indexID          = indexID
            , @database_nameID       = databaseID
            , @database_nameName     = databaseName
            , @fragmentation    = fragmentation
            , @partitionNumber  = partitionNumber
            , @pageCount        = page_count
        FROM #indexDefragList
        WHERE defragStatus = 0
        ORDER BY fragmentation DESC;

 
        -- Debug line
        IF (@debug_mode = 1 )
            RAISERROR('Step 5B - Looking up the specifics for our index ... ', 0, 42) WITH NOWAIT;
 
        /* Look up index information */
        SELECT @updateSQL = 
            ' UPDATE idl ' + 
            ' SET schemaName = QuoteName(s.name), ' + 
            '     objectName = QuoteName(o.name), ' + 
            '     indexName = QuoteName(i.name) ' + 
            ' FROM #indexDefragList AS idl ' + 
            '   INNER JOIN ' + @database_nameName + '.sys.objects AS o ON idl.objectID = o.object_id ' + 
            '   INNER JOIN ' + @database_nameName + '.sys.indexes AS i ON o.object_id = i.object_id ' + 
            '   INNER JOIN ' + @database_nameName + '.sys.schemas AS s ON o.schema_id = s.schema_id ' + 
            ' WHERE o.object_id = ' + CAST(@objectID AS VARCHAR(10)) + 
            '   AND i.index_id = ' + CAST(@indexID AS VARCHAR(10)) + 
            '   AND i.type > 0 ' + 
            '   AND idl.databaseID = ' + CAST(@database_nameID AS VARCHAR(10)) ;
        EXECUTE SP_EXECUTESQL @updateSQL;
 
        /* Grab our object names */
        SELECT 
            @objectName = objectName
            , @schemaName = schemaName
            , @indexName = indexName
        FROM 
            #indexDefragList
        WHERE 
            objectID = @objectID
            And indexID = @indexID
            And databaseID = @database_nameID;


        -- Debug line
        IF (@debug_mode = 1)
            RAISERROR('Step 5C - Grabbing the partition count ... ', 0, 42) WITH NOWAIT;
 
        /* Determine if the index is partitioned */
        SELECT @partitionSQL = 
            ' Select @partitionCount_OUT = Count(*) ' +
            ' From ' + @database_nameName + '.sys.partitions ' + 
            ' Where object_id = ' + CAST(@objectID AS VARCHAR(10)) + 
            ' And index_id = ' + CAST(@indexID AS VARCHAR(10)) + ';'
            , @partitionSQL_Param = '@partitionCount_OUT INT OUTPUT';
        EXECUTE SP_EXECUTESQL @partitionSQL, @partitionSQL_Param, @partitionCount_OUT = @partitionCount OUTPUT;


        -- Debug line
        IF (@debug_mode = 1 )
            RAISERROR('Step 5D - Seeing if there''s any LOBs to be handled...', 0, 42) WITH NOWAIT;
 
        /* Determine if the table contains LOBs */
        SELECT 
            @LOB_SQL = ' Select Top 1 @containsLOB_OUT = column_id ' + 
            ' From ' + @database_nameName + '.sys.columns With (NoLock) ' +  
            ' Where [object_id] = ' + CAST(@objectID AS VARCHAR(10)) + ' '  +
            ' And (system_type_id In (34, 35, 99) Or max_length = -1);',
            
            /*  system_type_id --> 34 = image, 35 = text, 99 = ntext
                max_length = -1 --> varbinary(max), varchar(max), nvarchar(max), xml */
            @LOB_SQL_Param = '@containsLOB_OUT int OutPut';
 
        EXECUTE SP_EXECUTESQL @LOB_SQL, @LOB_SQL_Param, @containsLOB_OUT = @containsLOB OUTPUT;
 
 
        -- Debug line
        IF (@debug_mode = 1)
            RAISERROR('Step 5E - Building our SQL statements ... ', 0, 42) WITH NOWAIT;
 
        /* If there's not a lot of fragmentation, or if we have a LOB, we should reorganize */
        IF (@fragmentation < @rebuild_pt) Or @containsLOB = 1 Or @partitionCount > 1
        BEGIN
 
            SET @sqlCommand = 
                'ALTER INDEX ' + @indexName + ' ON ' + @database_nameName + '.' +
                @schemaName + '.' + @objectName + ' REORGANIZE';
 
            /* If our index is partitioned, we should always reorganize */
            IF @partitionCount > 1
                SET @sqlCommand = @sqlCommand + 
                ' PARTITION = ' + CAST(@partitionNumber AS NVARCHAR(10));
 
        END;
 
        /* If the index is heavily fragmented and doesn't contain any partitions or LOB's, rebuild it */
        IF (@fragmentation >= @rebuild_pt) And IsNull(@containsLOB, 0) != 1 And @partitionCount <= 1
        BEGIN
 
            /* Set online rebuild options; requires Enterprise Edition */
            IF @online_rebuild = 1 And @editionCheck = 1 
                SET @rebuildCommand = ' REBUILD WITH (ONLINE = ON';
            ELSE
                SET @rebuildCommand = ' REBUILD WITH (ONLINE = OFF';
 
            /* Set processor restriction options; requires Enterprise Edition */
            IF (@MaxDop_Restriction IS NOT NULL) And @editionCheck = 1
                SET @rebuildCommand = @rebuildCommand + ', MAXDOP = ' + CAST(@MaxDop_Restriction AS VARCHAR(2)) + ')';
            ELSE
                SET @rebuildCommand = @rebuildCommand + ')';
 
            -- The final command
            SET @sqlCommand = 
                'ALTER INDEX ' + @indexName + ' ON ' + @database_nameName + '.' + 
                @schemaName + '.' + @objectName + @rebuildCommand;
 
        END;
 
        /* Are we executing the SQL?  If so, do it */
        IF @report_only = 0
        BEGIN
 
            -- Debug line
            IF (@debug_mode = 1)
                RAISERROR('Step 5F - Executing SQL statements...', 0, 42) WITH NOWAIT;
 
            /* Grab the time for logging purposes */
            SET @dateTimeStart  = GETDATE();
            EXECUTE SP_EXECUTESQL @sqlCommand;
            SET @dateTimeEnd  = GETDATE();
 
            /* Log our actions */
            INSERT INTO dbo.tbl_Maintain_Indexes
            (
                  databaseID
                , databaseName
                , objectID
                , objectName
                , indexID
                , indexName
                , partitionNumber
                , fragmentation
                , page_count
                , dateTimeStart
                , durationSeconds
            )
            SELECT
                  @database_nameID
                , @database_nameName
                , @objectID
                , @objectName
                , @indexID
                , @indexName
                , @partitionNumber
                , @fragmentation
                , @pageCount
                , @dateTimeStart
                , DATEDIFF(SECOND, @dateTimeStart, @dateTimeEnd);
 
            /* Just a little breather for the server */
            WAITFOR DELAY @defrag_delay;
 
            /* Print if specified to do so */
            IF (@print_commands = 1)
                PRINT 'Executed: ' + @sqlCommand;
        END
        ELSE
        
        /* Looks like we're not executing, just printing the commands */
        BEGIN
            -- Debug line
            IF (@debug_mode = 1)
                RAISERROR('Step 5F - Printing SQL statements...', 0, 42) WITH NOWAIT;
 
            -- Nothing to do
            IF @print_commands = 1 PRINT IsNull(@sqlCommand, 'error!');
        END
 
 
        -- Debug line
        IF (@debug_mode = 1)
            RAISERROR('Step 5G - Updating our index defrag status...', 0, 42) WITH NOWAIT;
 
        /* Update our index defrag list so we know we've finished with that index */
        UPDATE #indexDefragList
        SET defragStatus = 1
        WHERE 
          databaseID = @database_nameID
          And objectID = @objectID
          And indexID = @indexID
          And partitionNumber = @partitionNumber;
 
    END
 
    /* Do we want to output our fragmentation results? */
    IF @print_fragmentation = 1
    BEGIN
 
        -- Debug line
        IF @debug_mode = 1 
            RAISERROR('Step 5H - Displaying fragmentation results...', 0, 42) WITH NOWAIT;
 
        SELECT databaseID
            , databaseName
            , objectID
            , objectName
            , indexID
            , indexName
            , fragmentation
            , page_count
        FROM #indexDefragList;
 
    END;
 
    /* When everything is said and done, make sure to get rid of our temp table */
    DROP TABLE #indexDefragList;
    DROP TABLE #databaseList;
    DROP TABLE #processor;
 
 
    -- Debug line
    IF (@debug_mode = 1)
         RAISERROR('Finished running usp_Maintain_Indexes ...', 0, 42) WITH NOWAIT;

    SET NOCOUNT OFF;
	RETURN 0
END

GO